@@ -35,15 +35,11 @@ The gradient of all of the outputs with respect to the wavespeed, the
 source amplitudes, and the initial wavefields, may be calculated.
 """
 
-from typing import Optional, Union, List
+from typing import Optional, Union, List, Tuple
 import torch
 from torch import Tensor
-from deepwave.common import (set_dx, check_inputs, pad_model, pad_locations,
-                             location_to_index,
-                             extract_survey,
-                             get_n_batch, upsample, downsample,
-                             convert_to_contiguous, setup_pml,
-                             cfl_condition)
+from deepwave.common import (setup_propagator,
+                             downsample_and_movedim)
 
 
 class Scalar(torch.nn.Module):
@@ -71,7 +67,10 @@ class Scalar(torch.nn.Module):
             data should be stored to calculate the gradient with respect
             to `v` during backpropagation. Default False.
     """
-    def __init__(self, v, grid_spacing, v_requires_grad=False):
+    def __init__(self, v: Tensor,
+                 grid_spacing: Union[int, float, List[float],
+                                     Tensor],
+                 v_requires_grad: bool = False) -> None:
         super().__init__()
         if v.ndim != 2:
             raise RuntimeError("v must have two dimensions")
@@ -84,15 +83,23 @@ class Scalar(torch.nn.Module):
                 accuracy: int = 4, pml_width: Union[int, List[int]] = 20,
                 pml_freq: Optional[float] = None,
                 max_vel: Optional[float] = None,
-                survey_pad: Optional[Union[int, List[Optional[int]]]] = None,
+                survey_pad: Optional[Union[int,
+                                           List[Optional[int]]]] = None,
                 wavefield_0: Optional[Tensor] = None,
                 wavefield_m1: Optional[Tensor] = None,
-                psix_m1: Optional[Tensor] = None,
                 psiy_m1: Optional[Tensor] = None,
-                zetax_m1: Optional[Tensor] = None,
+                psix_m1: Optional[Tensor] = None,
                 zetay_m1: Optional[Tensor] = None,
-                origin: Optional[List[int]] = None, nt: Optional[int] = None,
-                model_gradient_sampling_interval: int = 1):
+                zetax_m1: Optional[Tensor] = None,
+                origin: Optional[List[int]] = None,
+                nt: Optional[int] = None,
+                model_gradient_sampling_interval: int = 1) -> Tuple[Tensor,
+                                                                    Tensor,
+                                                                    Tensor,
+                                                                    Tensor,
+                                                                    Tensor,
+                                                                    Tensor,
+                                                                    Tensor]:
         """Perform forward propagation/modelling.
 
         The inputs are the same as for :func:`scalar` except that `v` and
@@ -108,15 +115,16 @@ class Scalar(torch.nn.Module):
                       max_vel=max_vel,
                       survey_pad=survey_pad, wavefield_0=wavefield_0,
                       wavefield_m1=wavefield_m1,
-                      psix_m1=psix_m1, psiy_m1=psiy_m1,
-                      zetax_m1=zetax_m1, zetay_m1=zetay_m1,
+                      psiy_m1=psiy_m1, psix_m1=psix_m1,
+                      zetay_m1=zetay_m1, zetax_m1=zetax_m1,
                       origin=origin, nt=nt,
                       model_gradient_sampling_interval=
                       model_gradient_sampling_interval)
 
 
 def scalar(v: Tensor,
-           grid_spacing: Union[int, float, List[Union[int, float]], Tensor],
+           grid_spacing: Union[int, float, List[float],
+                               Tensor],
            dt: float,
            source_amplitudes: Optional[Tensor] = None,
            source_locations: Optional[Tensor] = None,
@@ -127,12 +135,15 @@ def scalar(v: Tensor,
            survey_pad: Optional[Union[int, List[Optional[int]]]] = None,
            wavefield_0: Optional[Tensor] = None,
            wavefield_m1: Optional[Tensor] = None,
-           psix_m1: Optional[Tensor] = None,
            psiy_m1: Optional[Tensor] = None,
-           zetax_m1: Optional[Tensor] = None,
+           psix_m1: Optional[Tensor] = None,
            zetay_m1: Optional[Tensor] = None,
+           zetax_m1: Optional[Tensor] = None,
            origin: Optional[List[int]] = None, nt: Optional[int] = None,
-           model_gradient_sampling_interval: int = 1):
+           model_gradient_sampling_interval: int = 1) -> Tuple[Tensor, Tensor,
+                                                               Tensor, Tensor,
+                                                               Tensor, Tensor,
+                                                               Tensor]:
     """Scalar wave propagation (functional interface).
 
     This function performs forward modelling with the scalar wave equation.
@@ -273,7 +284,7 @@ def scalar(v: Tensor,
         wavefield_m1:
             A Tensor specifying the initial wavefield at timestep -1. See
             the entry for `wavefield_0` for more details.
-        psix_m1, psiy_m1, zetax_m1, zetay_m1:
+        psiy_m1, psix_m1, zetay_m1, zetax_m1:
             Tensor specifying the initial value for this PML-related
             wavefield at timestep -1. See the entry for `wavefield_0`
             for more details.
@@ -308,7 +319,7 @@ def scalar(v: Tensor,
                 A Tensor containing the wavefield at timestep `nt`.
             wavefield_ntm1:
                 A Tensor containing the wavefield at timestep `nt-1`.
-            psix_ntm1, psiy_ntm1, zetax_ntm1, zetay_ntm1:
+            psiy_ntm1, psix_ntm1, zetay_ntm1, zetax_ntm1:
                 Tensor containing the wavefield related to the PML at timestep
                 `nt-1`.
             receiver_amplitudes:
@@ -318,96 +329,32 @@ def scalar(v: Tensor,
                 this Tensor will be empty.
 
     """
+    empty_string_list: List[str] = []
+    (v, other_models, source_amplitudes, wavefields,
+     ay, ax, by, bx, sources_i, receivers_i,
+     dy, dx, dt, nt, n_batch,
+     step_ratio, model_gradient_sampling_interval,
+     accuracy, pml_width_list) = \
+        setup_propagator(v, grid_spacing, dt, [], empty_string_list,
+                         [wavefield_0, wavefield_m1, psiy_m1, psix_m1,
+                          zetay_m1, zetax_m1],
+                         source_amplitudes,
+                         source_locations, receiver_locations,
+                         accuracy, pml_width, pml_freq, max_vel,
+                         survey_pad,
+                         origin, nt, model_gradient_sampling_interval)
 
-    # Check inputs
-    check_inputs(source_amplitudes, source_locations, receiver_locations,
-                 [wavefield_0, wavefield_m1, psix_m1, psiy_m1,
-                  zetax_m1, zetay_m1], accuracy, nt, v)
-
-    if nt is None:
-        nt = 0
-        if source_amplitudes is not None:
-            nt = source_amplitudes.shape[-1]
-    device = v.device
-    dtype = v.dtype
-    dx, dy = set_dx(grid_spacing)
-    if isinstance(pml_width, int):
-        pml_width = [pml_width for _ in range(4)]
-    fd_pad = accuracy // 2
-    pad = [fd_pad + width for width in pml_width]
-    models, locations = extract_survey(
-        [v], [source_locations, receiver_locations], survey_pad,
-        [wavefield_0, wavefield_m1, psix_m1, psiy_m1, zetax_m1, zetay_m1],
-        origin, pml_width
-    )
-    v, = models
-    source_locations, receiver_locations = locations
-    v_pad = pad_model(v, pad)
-    if max_vel is None:
-        max_vel = v.abs().max().item()
-    max_vel = abs(max_vel)
-    dt, step_ratio = cfl_condition(dx, dy, dt, max_vel)
-    if source_amplitudes is not None and source_locations is not None:
-        source_locations = pad_locations(source_locations, pad)
-        sources_i = location_to_index(source_locations, v_pad.shape[1])
-        source_amplitudes = (
-            -source_amplitudes * v_pad[source_locations[..., 0],
-                                       source_locations[..., 1],
-                                       None] ** 2 * dt**2
+    wfc, wfp, psiy, psix, zetay, zetax, receiver_amplitudes = \
+        torch.ops.deepwave.scalar(
+            v, source_amplitudes, wavefields[0], wavefields[1],
+            wavefields[2], wavefields[3], wavefields[4], wavefields[5],
+            ay, ax, by, bx, sources_i, receivers_i, dy, dx, dt, nt,
+            n_batch, step_ratio * model_gradient_sampling_interval,
+            accuracy, pml_width_list[0], pml_width_list[1], pml_width_list[2],
+            pml_width_list[3]
         )
-        source_amplitudes = upsample(source_amplitudes, step_ratio)
-    else:
-        sources_i = None
-    if receiver_locations is not None:
-        receiver_locations = pad_locations(receiver_locations, pad)
-        receivers_i = location_to_index(receiver_locations, v_pad.shape[1])
-    else:
-        receivers_i = None
-    n_batch = get_n_batch([source_locations, receiver_locations,
-                           wavefield_0, wavefield_m1, psix_m1, psiy_m1,
-                           zetax_m1, zetay_m1])
-    ax, ay, bx, by = \
-        setup_pml(pml_width, fd_pad, dt, v_pad, max_vel, pml_freq)
-    nt *= step_ratio
 
-    if source_amplitudes is not None:
-        if source_amplitudes.device == torch.device('cpu'):
-            source_amplitudes = torch.movedim(source_amplitudes, -1, 1)
-        else:
-            source_amplitudes = torch.movedim(source_amplitudes, -1, 0)
+    receiver_amplitudes = downsample_and_movedim(receiver_amplitudes,
+                                                 step_ratio)
 
-    v_pad = convert_to_contiguous(v_pad)
-    source_amplitudes = (convert_to_contiguous(source_amplitudes)
-                         .to(dtype).to(device))
-    wavefield_0 = convert_to_contiguous(wavefield_0)
-    wavefield_m1 = convert_to_contiguous(wavefield_m1)
-    psix_m1 = convert_to_contiguous(psix_m1)
-    psiy_m1 = convert_to_contiguous(psiy_m1)
-    zetax_m1 = convert_to_contiguous(zetax_m1)
-    zetay_m1 = convert_to_contiguous(zetay_m1)
-    sources_i = convert_to_contiguous(sources_i)
-    receivers_i = convert_to_contiguous(receivers_i)
-    ax = convert_to_contiguous(ax)
-    ay = convert_to_contiguous(ay)
-    bx = convert_to_contiguous(bx)
-    by = convert_to_contiguous(by)
-
-    wfc, wfp, psix, psiy, zetax, zetay, receiver_amplitudes = \
-        torch.ops.deepwave.scalar(v_pad, source_amplitudes, wavefield_0,
-                                  wavefield_m1, psix_m1, psiy_m1,
-                                  zetax_m1, zetay_m1, ax, ay, bx, by,
-                                  sources_i.long(), receivers_i.long(),
-                                  dx, dy, dt, nt, n_batch,
-                                  step_ratio *
-                                  model_gradient_sampling_interval,
-                                  accuracy, pml_width[0], pml_width[1],
-                                  pml_width[2], pml_width[3])
-
-    if receiver_amplitudes.numel() > 0:
-        if source_amplitudes.device == torch.device('cpu'):
-            receiver_amplitudes = torch.movedim(receiver_amplitudes, 1, -1)
-        else:
-            receiver_amplitudes = torch.movedim(receiver_amplitudes, 0, -1)
-        receiver_amplitudes = downsample(receiver_amplitudes, step_ratio)
-
-    return (wfc, wfp, psix, psiy, zetax, zetay, receiver_amplitudes)
+    return (wfc, wfp, psiy, psix, zetay, zetax, receiver_amplitudes)
